/*{{{  read me*/
/*
 * First attempt at pcl to sixel conversion
 *
 * I'm hoping that this file will be device-independent...???
 *
 * pcl allows sizes in points, dots or decipoints, or rows or columns
 * Internally, we use centipoints (a common multiple)
 *
 */
 
 
/*}}}*/

/*{{{  includes*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>

#define ITS_PCL_C  /* tell pcl.h it's us ! */
#include "pcl.h"
#undef ITS_PCL_C
/*}}}*/

/*{{{  sysdep stuff*/
#ifdef VMS
# define FOPEN(x,y) fopen(x,y,"mbc=60")
#else
# define FOPEN(x,y) fopen(x,y)
#endif
/*}}}*/

/*{{{  macros*/
#define ESCAPE 27


/* we keep several flags to keep track of what we have/not done
 * this is a simple shorthand for doing whatever is required to
 * bring a flag up to date
 */
 
#define ENSURE(flag, action)  do { if (!flag) { action ; flag=1; } } while(0) 
#define  CLEAR(flag, action)  do { if (flag)  { action ; flag=0; } } while(0) 
/*}}}*/

/*{{{  variables*/
static FILE *infile;


/*{{{  static INT32 factory_environment[ENVIRONMENT_SIZE] =*/
static INT32 factory_environment[ENVIRONMENT_SIZE] =
{
	0,             /* dummy */
	1,             /* copies */
	1,             /* internal tray */
	0,             /* A4 */
	0,             /* portrait  */
	0,   /* margins calculated at run time... */
	0,   /* right margin */
	0,   /* top margin */
	0,   /* bottom margin */
	1,             /* perf skip */

	1200,          /* HMI - 12 pt */
	1600,          /* VMI - 16 pt */
	0,             /* underline */

	75,            /* raster resolution */
	0,             /* raster margin */

	0,
	0,
	0,

	0,   /* macro */
	0,   /* line */
	0,   /* wrap */

	PRIMARY_FONT,   /* current font */
	0,  /* font id */
	0, /* char */
	
	/* primary */
	0,  /* symbol set */
	0,  /* id */
	0,  /* spacing */
	1200,  /* pitch */
	1200,  /* height */
	0,  /* style */
	0,  /* weight */
	0,  /* typeface */

	/* secondary */
	0,  /* symbol set */
	0,  /* id */
	0,  /* spacing */
	1200,  /* pitch */
	1200,  /* height */
	0,  /* style */
	0,  /* weight */
	0,  /* typeface */

};
/*}}}*/

static cursor_struct base_cursor = { NULL, 0, 0, 0 };
/*}}}*/

/*{{{  support routines*/
/*{{{  void fatal(char *fmt, ...)*/
void fatal(char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);

	exit(EXIT_FAILURE);
}
/*}}}*/

/*{{{  void to_stderr(char *fmt, ...)*/
void to_stderr(char *fmt, ...)
{
	va_list args;
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
}
/*}}}*/

/*{{{  void *xmalloc(int size)*/
void *xmalloc(int size)
{
	void *p = malloc(size);
	if (!p)
		fatal("Cant allocate memory\n");

	return p;
}
/*}}}*/

/*{{{  void *xrealloc(void *p, int size)*/
void *xrealloc(void *p, int size)
{
	/* bug in vaxcrtl */
	
	if (p)
		p = realloc(p, size);
	else
		p = malloc(size);
		
	if (!p)
		fatal("Cant allocate memory\n");

	return p;
}
/*}}}*/
/*}}}*/

/*{{{  pcl_ routines*/
/* many of these go on to call dev_* */

/*{{{  static void pcl_calc_margins()*/
static void pcl_calc_margins()
{
	if (environment[ENV_ORIENT])
	{
		environment[ENV_TOP] = UNPRINTABLE_LEFT;
		environment[ENV_LEFT] = UNPRINTABLE_TOP;
		environment[ENV_BOTTOM] = PAPER_WIDTH - UNPRINTABLE_RIGHT;
		environment[ENV_RIGHT] = PAPER_HEIGHT - UNPRINTABLE_BOTTOM;
	}
	else
	{
		environment[ENV_TOP] = UNPRINTABLE_TOP;
		environment[ENV_LEFT] = UNPRINTABLE_LEFT;
		environment[ENV_BOTTOM] = PAPER_HEIGHT - UNPRINTABLE_BOTTOM;
		environment[ENV_RIGHT] = PAPER_WIDTH - UNPRINTABLE_RIGHT;
	}

	DEBUG(("Calculated page size as "PERCENT_D32" * "PERCENT_D32"\n", PAPER_HEIGHT, PAPER_WIDTH));
	DEBUG(("Calculated top margin as "PERCENT_D32"\n", environment[ENV_TOP]));
	DEBUG(("Calculated left margin as "PERCENT_D32"\n", environment[ENV_LEFT]));
	DEBUG(("Calculated right margin as "PERCENT_D32"\n", environment[ENV_RIGHT]));
	DEBUG(("Calculated bottom margin as "PERCENT_D32"\n", environment[ENV_BOTTOM]));
 
}
/*}}}*/

/*{{{  static void pcl_reset(escape_struct *escape)*/
static void pcl_reset(escape_struct *escape)
{
	memcpy(environment, &factory_environment, sizeof(factory_environment));

	pcl_calc_margins();

	cursor = &base_cursor;
	cursor->valid = 0;
	cursor->x     = environment[ENV_LEFT];
	cursor->y     = environment[ENV_TOP];

}
/*}}}*/

/*{{{  static void pcl_init()*/
static void pcl_init()
{
	environment = xmalloc(sizeof factory_environment);
	pcl_reset(NULL);
}
/*}}}*/

/*{{{  static void pcl_close()*/
static void pcl_close()
{
	dev_close();
}
/*}}}*/

/*{{{  static void pcl_form_feed()*/
static void pcl_form_feed()
{
	ENSURE(virgin_page, dev_throw_page());
	cursor->valid=0;
	cursor->y=environment[ENV_TOP];
}
/*}}}*/

/*{{{  static void pcl_line_feed()*/
static void pcl_line_feed()
{
	cursor->y += environment[ENV_VMI];
	cursor->valid = 0;
	if (environment[ENV_PERF] && cursor->y >= environment[ENV_BOTTOM])
		pcl_form_feed();
}
/*}}}*/

/*{{{  static void pcl_page_setup()*/
static void pcl_page_setup()
{
	pcl_form_feed();
	pcl_calc_margins();
	environment[ENV_HMI] = factory_environment[ENV_HMI];
	environment[ENV_VMI] = factory_environment[ENV_VMI];
}
/*}}}*/
	
/*{{{  static void pcl_change_env(env_offsets element, INT32 value)*/
static void pcl_change_env(env_offsets element, INT32 value)
{
	switch (element)
	{
	}
	
	if (value != environment[element])
	{	DEBUG(("Change environment["PERCENT_D32"] from " PERCENT_D32 " to " PERCENT_D32 "\n",
	         element, environment[element], value));
	         
		environment[element] = value;
		dev_env_changed(element);
	}
}
/*}}}*/

/*{{{  void pcl_area_fill(escape_struct *escape)*/
void pcl_area_fill(escape_struct *escape)
{
	static int warned = 0;

	if (!warned)
		WARN(("pcl_area_fill not implemented yet\n"));

	warned=1;
}
/*}}}*/

/*{{{  void pcl_handle_control_char(int ch)*/
void pcl_handle_control_char(int ch)
{
	switch (ch)
	{
		case 0x0a :
			pcl_line_feed();
			if (environment[ENV_LINE] & 1)
				cursor->x = environment[ENV_LEFT];
			break;

		case 0x0d :
			cursor->x = environment[ENV_LEFT];
			if (environment[ENV_LINE] & 1)
				pcl_line_feed();
			break;

		case 0x0c :
			ENSURE(virgin_page, dev_throw_page());
			cursor->valid=0;
			cursor->y=environment[ENV_TOP];
			if (environment[ENV_LINE & 2])
				cursor->x = environment[ENV_LEFT];
			break;
			
		default :
			WARN(("Ignoring character 0x%x\n", ch));
	}
}
/*}}}*/


/*{{{  static void pcl_begin_raster(escape_struct *escape)*/
static void pcl_begin_raster(escape_struct *escape)
{
	if (escape->param)
		environment[ENV_RASTER_MARGIN] = cursor->x;
	else
		environment[ENV_RASTER_MARGIN] = environment[ENV_LEFT];

	dev_begin_raster(escape);
}
/*}}}*/

/*{{{  static void pcl_raster_data(escape_struct *escape)*/
static void pcl_raster_data(escape_struct *escape)
{
	dev_raster_data(escape);
	cursor->y += 7200 / environment[ENV_RASTER_RES];
}
/*}}}*/

	
/*{{{  static void pcl_push_pop(escape_struct *escape)*/
static void pcl_push_pop(escape_struct *escape)
{
	if (escape->param)
		/*{{{  pop*/
		{	
			cursor_struct *next = cursor->next;
			if (next)
			{	free(cursor);
				cursor = next;
			}
			else
				DEBUG(("Cant pop a cursor\n"));
		}	
		/*}}}*/
	else
		/*{{{  push*/
		{
			cursor_struct *new = xmalloc(sizeof(cursor_struct));
		
			new->next = cursor;
			new->x    = cursor->x;
			new->y    = cursor->y;
			cursor    = new;
		}
		/*}}}*/
} 
/*}}}*/

/*{{{  static void pcl_cursor_x(escape_struct *escape)*/
static void pcl_cursor_x(escape_struct *escape)
{
	switch(*escape->pointer)
	{
		case '+' : /* for -ve, param is -ve already */
		case '-' : cursor->x += escape->param;
		           break;
		default  : cursor->x = escape->param;
	}

	if (cursor->x < environment[ENV_LEFT])
		cursor->x = environment[ENV_LEFT];

	if (cursor->x > environment[ENV_RIGHT])
		cursor->x = environment[ENV_RIGHT];
}
/*}}}*/

/*{{{  static void pcl_cursor_y(escape_struct *escape)*/
static void pcl_cursor_y(escape_struct *escape)
{
	switch(*escape->pointer)
	{
		case '+' : /* for -ve, param is -ve already */
		case '-' : cursor->y += escape->param;
		           break;
		default  : cursor->y = escape->param;
	}

	if (cursor->y < 0)
		cursor->y = 0;
}
/*}}}*/

/*{{{  static void pcl_line_space(escape_struct *escape)*/
static void pcl_line_space(escape_struct *escape)
{
	/* handle line spacing sequence */
	
	if (escape->param)
		pcl_change_env(ENV_VMI, 7200/escape->param);
}
/*}}}*/

/*{{{  static void pcl_pitch(escape_struct *escape)*/
static void pcl_pitch(escape_struct *escape)
{
	int offset = escape->string[0]=='(' ? PRIMARY_FONT : SECOND_FONT;
	pcl_change_env(offset + ENV_PRIM_PITCH,(INT32) (7200 / escape->real));
}
/*}}}*/

/*{{{  static void pcl_do_bM(escape_struct *escape)*/
static void pcl_do_compression(escape_struct *escape)
{
	if (escape->param)
		WARN(("Raster compression not supported\n"));
}
/*}}}*/

/*}}}*/

/*{{{  short escape table*/
escape_entry_struct short_escape_table[] =
 /* sequence  desc             fn to call     environment  properties scale */
{
	{ "E",    "Reset",          dev_reset,       0,         PAGE_SETUP | INVALIDATES_CURSOR | INVALIDATES_FONT},
	{"",      NULL }
};
/*}}}*/

/*{{{  long escape table*/
escape_entry_struct long_escape_table[] =
 /* sequence  desc             fn to call      environment  properties scale*/
{
	{ "&lX", ""PERCENT_D32" Copies",        NULL,             ENV_COPIES },
	{ "&lH", "Paper source "PERCENT_D32"",  NULL,             ENV_SOURCE,    FLUSH_PAGE },
	{ "&lA", "Page size "PERCENT_D32"",     NULL,             ENV_PAGE_SIZE, PAGE_SETUP },
/*	{ "&lP", "Page length "PERCENT_D32"",   NULL,             ENV_PAGE_LENGTH, PAGE_SETUP, SCALE_ROWS },*/
	{ "&kH", "HMI ["PERCENT_D32"]",         NULL,             ENV_HMI,   0, SCALE_120 },
	{ "&lC", "VMI ["PERCENT_D32"]",         NULL,             ENV_VMI,   0, SCALE_48 },
	{ "&lD", "lines per inch",   pcl_line_space },
	{ "&aL", "left margin "PERCENT_D32"",   NULL,             ENV_LEFT,  0, SCALE_COLS },
	{ "&aM", "right margin "PERCENT_D32"",  NULL,             ENV_RIGHT, 0, SCALE_COLS },
	{ "&lE", "top margin "PERCENT_D32"",    NULL,             ENV_TOP,   0 , SCALE_ROWS },
	{ "&lL", "Perforation "PERCENT_D32"",   NULL,             ENV_PERF },
	{ "&lO", "orientation "PERCENT_D32"",   NULL,             ENV_ORIENT,PAGE_SETUP },

	{ "&aC", "Cursor x column ["PERCENT_D32"]",  pcl_cursor_x,     0,         INVALIDATES_CURSOR, SCALE_COLS },
	{ "&aH", "Cursor x deci   ["PERCENT_D32"]",    pcl_cursor_x,     0,         INVALIDATES_CURSOR, SCALE_DECIPOINT},
	{ "*pX", "Cursor x dot    ["PERCENT_D32"]",     pcl_cursor_x,     0,         INVALIDATES_CURSOR, SCALE_DOT },
	{ "&aR", "Cursor y row    ["PERCENT_D32"]",     pcl_cursor_y,     0,         INVALIDATES_CURSOR, SCALE_ROWS },
	{ "&aV", "Cursor y deci   ["PERCENT_D32"]",    pcl_cursor_y,     0,         INVALIDATES_CURSOR, SCALE_DECIPOINT},
	{ "*pY", "Cursor y dot    ["PERCENT_D32"]",     pcl_cursor_y,     0,         INVALIDATES_CURSOR, SCALE_DOT },
	{ "&fS", "Push/pop cursor",  pcl_push_pop,     0,         INVALIDATES_CURSOR},

	{ "*tR", "Raster Res ["PERCENT_D32"]",       NULL,             ENV_RASTER_RES },
	{ "*rA", "Begin Raster ["PERCENT_D32"]",     pcl_begin_raster, 0,  NEEDS_CURSOR },
	{ "*bW", "Raster data ["PERCENT_D32"]", pcl_raster_data,  0,  MARKS_PAGE | EXPECTS_DATA },
	{ "*rB", "End raster",       dev_end_raster,   0,  INVALIDATES_CURSOR},

	{ "*cH", "Rectangle width",  NULL,             ENV_WIDTH,  0, SCALE_DECIPOINT },
	{ "*cA", "Rectangle width",  NULL,             ENV_WIDTH,  0, SCALE_DOT },
	{ "*cV", "Rectangle height", NULL,             ENV_HEIGHT, 0, SCALE_DECIPOINT },
	{ "*cB", "Rectangle height", NULL,             ENV_HEIGHT, 0, SCALE_DOT },
	{ "*cG", "Area fill ID",     NULL,             ENV_AREA },
	{ "*cP",  "Fill area",       dev_area_fill,    0,          NEEDS_CURSOR | MARKS_PAGE },
	
	{ ")sW", "Font descriptor",  dev_font_descriptor, 0, EXPECTS_DATA },
	{ "(sW", "Char. descriptor", dev_char_descriptor, 0, EXPECTS_DATA },
	{ "*cF", "Font control",     dev_font_control },
	{ "*cD", "Font ID",          NULL,             ENV_FONT_ID },
	{ "*cE", "Character code",   NULL,             ENV_CHAR },

	{ "( X", "primary id",       NULL,             ENV_PRIM_ID , INVALIDATES_FONT},
	{ ") X", "secondary id",     NULL,             ENV_SCND_ID , INVALIDATES_FONT},
	
	{ "(sP", "primary spacing",  NULL,             ENV_PRIM_SPACING , INVALIDATES_FONT},
	{ ")sP", "secondary spacing",NULL,             ENV_SCND_SPACING , INVALIDATES_FONT},
	{ "(sH", "primary pitch",    pcl_pitch,        0, INVALIDATES_FONT},
	{ ")sH", "secondary pitch",  pcl_pitch,        0, INVALIDATES_FONT},
	{ "(sV", "primary height",   NULL,             ENV_PRIM_HEIGHT , INVALIDATES_FONT, SCALE_POINT},
	{ ")sV", "secondary height", NULL,             ENV_SCND_HEIGHT , INVALIDATES_FONT, SCALE_POINT},
	{ "(sS", "primary style",    NULL,             ENV_PRIM_STYLE , INVALIDATES_FONT},
	{ ")sS", "secondary style",  NULL,             ENV_SCND_STYLE , INVALIDATES_FONT},
	{ "(sB", "primary stroke",   NULL,             ENV_PRIM_WEIGHT , INVALIDATES_FONT},
	{ ")sB", "secondary stroke", NULL,             ENV_SCND_WEIGHT , INVALIDATES_FONT},
	{ "(sT", "primary typeface", NULL,             ENV_PRIM_TYPEFACE , INVALIDATES_FONT},
	{ ")sT", "secondary typeface",NULL,            ENV_SCND_TYPEFACE , INVALIDATES_FONT},

	{ "*bM", "Compression mode ?", pcl_do_compression },
	
	{ "",    NULL }
};

/*}}}*/

/*{{{  escape parsing*/
/*{{{  read me*/
/* pcl escape sequences come in two flavours...
 *
 * ESC X   - X between 48 and 126
 *
 * eg ESC E = reset
 *
 *
 * ESC X y # z1 # z2 # z3 ... # Zn
 *
 * X in 33 - 47  (! - /)
 * y = (optional ?) group 96-126
 * # = optional value field - omitted => 0
 *                            might have + or -
 *                            can have decimal point
 *
 * z = param, 96-126
 * Z = param  64-94
 *
 * above sequence is really just a condensed form of
 * ESC X y # Z1
 * ESC X y # Z2
 * ESC X y # Z3
 * ...
 * ESC X y # Zn
 *
 * eg ESC & l 1 o 2 A
 * ESC & l 1 O =  (something)
 * ESC & l 2 A =  (something else)
 *
 *
 *
 * Some sequences mean interpret the following data in a certain way
 * eg raster graphics. This sequence can be in the middle of a string,
 * so we must read in the entire string before processing it.
 *
 * We read in any block data ourselves, in case the device chooses not
 * to implement the fn.
 *
 * Rather than continually mallocing, we keep one block, and realloc
 * it as necessary
 */
/*}}}*/

static unsigned char *escape_data_buffer = NULL;
static int escape_data_size = 0;

/*{{{  static void lookup_escape(escape_struct *escape, escape_entry_struct *p)*/
static void lookup_escape(escape_struct *escape, escape_entry_struct *p)
{
	/*{{{  lookup command in table*/
	for ( ; p->desc ; ++p)
		if (strcmp(escape->string, p->command)==0)
			goto found;
			
	/*{{{  get here => not found in table*/
	WARN(("Unknown escape sequence %s\n", escape->string));
	return;
	/*}}}*/
	
	found:
	/*}}}*/

	escape->entry = p;
	
	/*{{{  scale the param to centipoints*/
	switch (p->scaling_factor)
	{
		case SCALE_ROWS : escape->param = (INT32) (escape->real * environment[ENV_HMI]); break;
		case SCALE_COLS : escape->param = (INT32) (escape->real * environment[ENV_VMI]); break;
		case 0 :          escape->param = (INT32) escape->real; break;
		default :         escape->param = (INT32) (escape->real * p->scaling_factor); break;
	}
	/*}}}*/

	DEBUG(("esc-%s : ", p->command));
	DEBUG((p->desc, escape->param));
	DEBUG(("\n"));

	if (p->properties & FLUSH_PAGE)
		pcl_form_feed();

	if (p->properties & PAGE_SETUP)
		pcl_page_setup();
		
	if (p->properties & EXPECTS_DATA)
		/*{{{  read it in*/
		{
			if (escape_data_size < escape->param)
				escape_data_buffer =
				   xrealloc(escape_data_buffer, escape_data_size = escape->param);
				
			escape->data = escape_data_buffer;
			
			if (fread(escape->data, 1, escape->param, infile) < escape->param)
				fatal("fread read fewer bytes than asked for !\n");
		}
		/*}}}*/
	
	if (p->environment)
		/*{{{  make the change, and tell the device*/
			pcl_change_env(p->environment, escape->param);  /* tells the device */
		/*}}}*/

	if (p->properties & NEEDS_CURSOR)
		ENSURE(cursor->valid, dev_set_cursor(escape) );
	
	if (p->properties & MARKS_PAGE)
		CLEAR(virgin_page, dev_about_to_mark_page());

	if (p->action)
		p->action(escape);

	if (p->properties & INVALIDATES_CURSOR)
		cursor->valid = 0;

	if (p->properties & INVALIDATES_FONT)
		/*{{{  mark font as invalid, unless aren't using modified font*/
		{
			if (  ( (escape->string[0]=='(') && (environment[ENV_CURRENT_FONT]==SECOND_FONT ))
				 ||( (escape->string[0]==')') && (environment[ENV_CURRENT_FONT]==PRIMARY_FONT))
			   )
			   /* nothing */ ;
			else
				font_valid = 0;
		}
		/*}}}*/
}
/*}}}*/

/*{{{  static void do_escape()*/
static void do_escape()
{
	escape_struct escape;

	escape.string[0] = fgetc(infile);

	if (escape.string[0] >= 48)
		/*{{{  two-char escape*/
		{
			escape.string[1] = 0;
		
			lookup_escape(&escape, short_escape_table);
		} 
		
		/*}}}*/
	else
		/*{{{  extended escape*/
		{
			static char escape_buffer[100];
			char *p = escape_buffer;
			int ch = fgetc(infile);
			
			escape.string[3]=0;
				
			/*{{{  get group letter*/
			/* y is optional - if not in range, put it in buffer and set it ' ' */
			
			if (ch >= 96 && ch <= 126)
			{	
				escape.string[1]=ch;
				ch = getc(infile);
			}
			else
				escape.string[1] = ' ';
			/*}}}*/
		
			/*{{{  fill escape buffer*/
			while ( (*p++ = ch) < 64 || ch > 94)
				ch = fgetc(infile);
			/*}}}*/
		
			*p = 0;
			
			p = escape_buffer;
		
			do
			{
				/*{{{  collect elements of command*/
				escape.pointer = p;
				
				/* cant use strtod straight off, as we might meet 123e456 */
				
				p += strspn(p, "+-.0123456789");
				
				ch	= *p;
				*p++ = 0;
				
				escape.real = strtod(escape.pointer, NULL);  /* real for generality */
				
				escape.string[2] = (ch & 0x5f);
				/*}}}*/
		
				if ( (escape.string[0]=='(' || escape.string[0]==')') && escape.string[1]==' ')
					/*{{{  it's a font symbol-set command*/
					WARN(("symbol-set selection not implemented\n"));
					/*}}}*/
				else
					lookup_escape(&escape, long_escape_table);
		
			} while(ch >= 96);
		}
		
		/*}}}*/
}
/*}}}*/
/*}}}*/

/*{{{  int main(int argc, char *argv[])*/
int main(int argc, char *argv[])
{	int ch;
	int in_a_string = 0;
	
	/*{{{  parse args*/
	infile = stdin;   /* vax doesn't like compile-time init */
	outfile = stdout;
	
	while (--argc && (++argv)[0][0]=='-')
		switch (argv[0][1])
		{	case 'v' : debug = 1; break;
			case 'q' : quiet = 1; break;
			case 'o' : factory_environment[ENV_ORIENT] = atoi(argv[0]+2); break;
			case 'l' : factory_environment[ENV_LINE] = atoi(argv[0]+2); break;
			case 'w' :factory_environment[ENV_WRAP] = atoi(argv[0]+2); break;
			case 'p' :factory_environment[ENV_PERF] = atoi(argv[0]+2); break;
			default  : fprintf(stderr, "Unknown option \"%s\" ignored\n", argv[0]);
		}
	
	if (argc && !(infile = FOPEN(argv[0], "rb")))
	{	fprintf(stderr, "Cant read file \"%s\"\n", argv[0]);
		exit(EXIT_FAILURE);
	}
	
	if (argc > 1 && !(outfile = FOPEN(argv[1], "w")))
	{	fprintf(stderr, "Cant write file \"%s\"\n", argv[1]);
		exit(EXIT_FAILURE);
	}
	/*}}}*/

	dev_init();
	pcl_init();

	while ((ch = fgetc(infile)) != EOF)
		if (isprint(ch))
			/*{{{  start a string, if necessary, and send char*/
			{
				CLEAR(virgin_page, dev_about_to_mark_page() );
				ENSURE(font_valid,    dev_set_font(environment[ENV_CURRENT_FONT]) );
				ENSURE(cursor->valid, dev_set_cursor(NULL)   );
				ENSURE(in_a_string,   dev_start_string() );
				
				dev_write_char(ch);
			
				cursor->x += environment[ENV_HMI];
			
				if (cursor->x >= environment[ENV_RIGHT] && environment[ENV_WRAP])
				{
					in_a_string = 0;
					dev_end_string();
					cursor->valid = 0;
					cursor->x = environment[ENV_LEFT];
					pcl_line_feed();
				}
				
			}
			/*}}}*/
		else
		{
			/*{{{  flush a string, if necessary*/
			CLEAR(in_a_string, dev_end_string());
			/*}}}*/
			if (ch == ESCAPE)
				do_escape();
			else
				pcl_handle_control_char(ch);
		}
				
	/*{{{  flush a string, if necessary*/
	CLEAR(in_a_string, dev_end_string());
	
	/*}}}*/
	
	ENSURE(virgin_page, dev_throw_page());
	pcl_close();
	
	return EXIT_SUCCESS;
}
/*}}}*/
