#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pcl.h"

/*{{{  readme*/
/*
 * pcl.c works in centipoints internally, so we use '0.01 0.01 scale'
 * to make that our units too.
 *
 * for raster data -
 *
 * really want to use the readhexstring method of image, ie
 *
 * /my_string 256 string def
 * ...
 * { currentfile my_string readhexstring pop } image
 *
 * since this saves on virtual memory - if we just define the image
 * as a constant string, it takes up virtual memory forever
 * (or until a restore)
 *
 * However, we dont know the size of the image until we've read it
 * in. So we just have to render one line at a time.
 *
 */
/*}}}*/

/*{{{  defines and prologue*/
#define NULL_FN(x) void x (escape_struct *escape) { }


/* postscript origin is bottom left, but pcl is top left... */
#define CURSOR_X (cursor->x)
#define CURSOR_Y ( ( environment[ENV_ORIENT] ? PAPER_WIDTH : PAPER_HEIGHT) - (cursor->y))

#define PROLOG               \
"%%!PS\n"                     \
"/pcldict 20 dict def\n"        \
"pcldict begin\n"                \
/* RASTER STUFF */                \
"/read_raster { currentfile rbuf readhexstring pop } def\n"           \
"/read_compressed { "                                                  \
" currentfile rc readhexstring pop"  /* read a pair into rc*/           \
" {} forall"  /* break into individual stack elements - count, value*/   \
" 0 1 3 index 1 sub "  /* set up 0 1 <count-1> for a for loop */          \
" { rbuf exch 2 index put } for "  /* put value into rbuf <count> times*/  \
" pop rbuf exch 0 exch getinterval "   /* make a substring length <count> */\
"} def\n"                                                 \
"/rc 2 string def\n"    /* read 2 hex digits at a time */  \
"/rl {"                                                     \
" rm exch 5 exch put"  /* put y translate into matrix */     \
" rw 1 true rm { read_raster } imagemask "                    \
" } def\n"                                                     \
"/rld { rm exch 5 exch put"  /* put y translate into matrix */  \
"       rw 1 true rm { read_compressed } imagemask } def\n"      \

/* rw and rbuf are defined once width is known */
/*}}}*/

/*{{{  variables*/
static int raster_width=0;

static int page_count=0;
/*}}}*/

/*{{{  void dev_init()*/
void dev_init()
{
	fprintf(outfile, PROLOG);
}
/*}}}*/
/*{{{  void dev_reset(escape_struct *escape)*/
void dev_reset(escape_struct *escape)
{ 
}
/*}}}*/
/*{{{  void dev_throw_page()*/
void dev_throw_page()
{
	fputs("showpage\n", outfile);
	++page_count;
}
/*}}}*/
/*{{{  void dev_close()*/
void dev_close()
{
	fprintf(outfile, "end\n");
}
/*}}}*/


/*{{{  void dev_about_to_mark_page()*/
void dev_about_to_mark_page()
{
	fprintf(outfile, "0.01 0.01 scale\n");
	if (environment[ENV_ORIENT])
		fprintf(outfile, "0 0 moveto 90 rotate 0 "PERCENT_D32" translate\n", -PAPER_WIDTH);
}
/*}}}*/
	
/*{{{  void dev_env_changed(env_offsets element)*/
void dev_env_changed(env_offsets element)
{
	switch (element)
	{
		/*{{{  case ENV_COPIES :*/
		case ENV_COPIES :
			fprintf(outfile, "/#copies "PERCENT_D32" def\n", environment[ENV_COPIES]);
			break;
		/*}}}*/
			
	}			
}
/*}}}*/

/*{{{  readme - raster stuff*/
/* as I understand it, the matrix is essentially how we map any point
 * on the page to a point in our image. Now my book recommends using
 * translate and scale to map page coords to a 1*1 square, then
 * maping 1*1 sqaure to image, but I dont fancy all that...
 * SO - X,Y = image coords (0->n, 0->1), x,y = page co-ors (centi-pts).
 * matrix  [ a b c d e f ]
 * X = ax+by+e  Y = cx+dy+f
 * straight away, b=c=0
 * raster_res => r pix/inch = r/7200 pix/cp => a = r/7200
 * ditto d => a=d=r/7200 [ check - double r => double a - correct ]
 * when x = raster margin, X = 0 => e=-a*raster_margin
 * when y = cursor->y,     Y = 0 => f=-d*cursor->y
 *
 * only the y translate should change for each line, so to save
 * VM, we prepare one copy of the matrix in advance
 * To save space in ps file, we hard-wire as many numbers into the procedure
 * defn as poss...
 *
 * Also, we define a procedure using run-length encoding, and for each
 * line, we use either run-length encoding or uncompressed coding
 *
 * ps variables :
 *  rbuf - one string into which lines are built or uncompressed
 *  rc   - 2-char string, for reading compressed data with readhexstring
 *  rl   - routine which renders a (uncompressed) line
 *       - on entry, y translation for matrix is on stack
 *  rld  - decompresses and renders a line
 */
/*}}}*/
/*{{{  raster defines*/
/*}}}*/
/*{{{  void dev_begin_raster(escape_struct *escape)*/
void dev_begin_raster(escape_struct *escape)
{
	double a =  environment[ENV_RASTER_RES] / 7200.0;

	raster_width = 0;  /* some things we cant do until first line */

  /* matrix (all but y translate) */
  
	fprintf(outfile, "/rm  [%lf 0 0 %lf %lf 0] def\n",
	      a, a, -a*environment[ENV_RASTER_MARGIN]);
}
/*}}}*/
/*{{{  void dev_raster_data(escape_struct *escape)*/
void dev_raster_data(escape_struct *escape)
{
	int i;
	int compressed_width = 0;
	int last_byte = -1;
	
	double a =  environment[ENV_RASTER_RES] / 7200.0;

	/*{{{  work out width if we compressed ; abandon if all 0's*/
	for (i=0; i<escape->param; ++i)
		if (escape->data[i] != last_byte)
		{	last_byte = escape->data[i];
			compressed_width += 2;
		}
	
	if (compressed_width == 2 && last_byte == 0)
		return;   /* it's a whole line of zero's */
	
	DEBUG(("Raw raster width "PERCENT_D32" - compressed width %d\n", escape->param, compressed_width));
	/*}}}*/

	/*{{{  check raster-width against escape param*/
	if (raster_width != escape->param)
	{
		if (raster_width)
			WARN(("Raster image is not a uniform width\n"));
	
		raster_width = escape->param;
			
		fprintf(outfile,
			"/rbuf %d string def\n/rw %d def\n",
		   raster_width, raster_width*8);
	}
	/*}}}*/

	if (compressed_width < raster_width)
		/*{{{  use compression*/
		{
			int count = 0;
			fprintf(outfile, "%.2lf rld ", -a*CURSOR_Y);
		
			last_byte=-1;
			
			for (i=0; i<raster_width; ++i)
			{
				if (escape->data[i] != last_byte)
				{
					while (count > 255)
					{	fprintf(outfile, "FF%02x", last_byte);
						count -= 255;
					}
					if (count)
						fprintf(outfile, "%02x%02x", count, last_byte);
					last_byte = escape->data[i];
					count = 1;
				}
				else
					++count;
			}
			
			while (count > 255)
			{	fprintf(outfile, "FF%02x", last_byte);
				count -= 255;
			}
			if (count)
				fprintf(outfile, "%02x%02x", count, last_byte);
		}
		
		/*}}}*/
	else
		/*{{{  dont*/
		{
			fprintf(outfile, "%.2lf rl ", -a*CURSOR_Y);
		
			for (i=0; i<raster_width; ++i)
				fprintf(outfile, "%02x", escape->data[i]);
		}
		
		/*}}}*/

	fputs("\n", outfile);
}
/*}}}*/
/*{{{  void dev_end_raster(escape_struct *escape)*/
void dev_end_raster(escape_struct *escape)
{
	cursor->valid = 0;  /* just in case */
}
/*}}}*/

/*{{{  void dev_set_cursor(escape_struct *escape)*/
void dev_set_cursor(escape_struct *escape)
{
	fprintf(outfile, PERCENT_D32 " " PERCENT_D32 " moveto ", CURSOR_X, CURSOR_Y);
}
/*}}}*/
/*{{{  void dev_set_font(int offset)*/
void dev_set_font(int offset)
{
	static int warned_spacing=0;
	INT32 *env = environment + offset;
	
	char *font;
	
	if (env[ENV_PRIM_SPACING]==0 && !warned_spacing)
	{	WARN(("ps only does proportional spacing\n"));
		warned_spacing = 1;
	}

	switch (env[ENV_PRIM_TYPEFACE])
	{
		case 3 : font = "Courier";     break;
		case 4 : font = "Helvetica";   break;
		case 5 : font = "Times-Roman"; break;
		case 6 : font = "Hershey-Gothic-English"; break;
		
		default : font = "Courier";     break;
	}
	
	fprintf(outfile, "/%s findfont "PERCENT_D32" scalefont setfont\n",
	   font, env[ENV_PRIM_HEIGHT]);
}
/*}}}*/

/*{{{  void dev_start_string()*/
void dev_start_string()
{
	fputc('(', outfile);
}
/*}}}*/
/*{{{  void dev_write_char(int ch)*/
void dev_write_char(int ch)
{
	if (ch=='\\' || ch=='(' || ch==')' )
		fputc('\\', outfile);
		
	fputc(ch, outfile);
}
/*}}}*/
/*{{{  void dev_end_string()*/
void dev_end_string()
{
	fputs(") show\n", outfile);
}
/*}}}*/

/*{{{  void dev_area_fill(escape_struct *escape)*/
void dev_area_fill(escape_struct *escape)
{
	pcl_area_fill(escape);
}
/*}}}*/


NULL_FN(dev_font_control)
NULL_FN(dev_select_font)
NULL_FN(dev_font_descriptor)
NULL_FN(dev_char_descriptor)
