#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pcl.h"

/*{{{  defines*/
#define CSI "\x1b["

#define SET_SSU_STRING CSI "11h" CSI "?1 I"

#define RASTER_WIDTH   8192   /* should be ample */

#define SCALE 24   /* internal sizes in centipoints, output in dots */

#define NULL_FN(x) void x (escape_struct *escape) { }

/*}}}*/

/*{{{  variables*/
/*}}}*/

/*{{{  raster stuff*/
static char raster_buffer[RASTER_WIDTH];

/* we have to read 6 lines of pcl graphics for each sixel strip.
 * mask records which row we've got to
 */
 
static int  raster_mask = 1;
static int raster_width = 0;  /* max we have seen */

/*{{{  static void empty_raster()*/
static void empty_raster()
{
	memset(raster_buffer, 0, sizeof raster_buffer);
	raster_mask = 1;
	raster_width= 0;
}
/*}}}*/

/* in order to minimise size of file, we use the !<count><sixel> packing */

/*{{{  FLUSH(count, sixel)*/
#define FLUSH(count,sixel)                      \
	if (count>3)                                  \
		fprintf(outfile, "!%d%c", count, sixel+63); \
	else                                            \
		while (count--)                               \
			fputc(sixel+63, outfile);
/*}}}*/

/*{{{  static void flush_raster()*/
static void flush_raster()
{	int x;
	char *sixel=raster_buffer;
	int last_sixel = -1;
	int sixel_count = 0;

	for (x=0; x<raster_width; ++x, ++sixel)
		if (*sixel == last_sixel)
			++sixel_count;
		else if (sixel_count)
		{	FLUSH(sixel_count, last_sixel)
			last_sixel=*sixel;
			sixel_count=1;
		}
		else  /* this is the first sixel of the line; dont worry ! */
		{	sixel_count=1;
			last_sixel=*sixel;
		}


	if ((last_sixel > 0) && sixel_count)
	{	FLUSH(sixel_count, last_sixel)
	}

	fputs("-\n", outfile);
	
	empty_raster();

}
/*}}}*/
/*}}}*/

/*{{{  void dev_init()*/
void dev_init()
{
	empty_raster();
	fprintf(outfile, CSI"!p" CSI"11h" CSI"7 I");
}
/*}}}*/
/*{{{  void dev_close()*/
void dev_close()
{
}

/*}}}*/
/*{{{  void dev_reset(escape_struct *escape)*/
void dev_reset(escape_struct *escape)
{ 
	fputs("\033c", outfile);
	empty_raster();
}
/*}}}*/
/*{{{  void dev_throw_page()*/
void dev_throw_page()
{
	fputc('\f', outfile);
}
/*}}}*/

/*{{{  void dev_about_to_mark_page()*/
void dev_about_to_mark_page()
{
}
/*}}}*/
/*{{{  void dev_env_changed(escape_struct *escape, env_offsets element)*/
void dev_env_changed(env_offsets element)
{
	switch (element)
	{
		/*{{{  case ENV_COPIES :*/
		case ENV_COPIES :
			fprintf(outfile, CSI"%d&q", environment[ENV_COPIES]);
			break;
		/*}}}*/
		/*{{{  case ENV_ORIENT :*/
		case ENV_ORIENT :
			fprintf(outfile, CSI"%d z", environment[ENV_ORIENT]+1);
			break;
/*}}}*/

	}			
}
/*}}}*/

/*{{{  void dev_begin_raster(escape_struct *escape)*/
void dev_begin_raster(escape_struct *escape)
{
	if (environment[ENV_RASTER_MARGIN]/SCALE != environment[ENV_LEFT]/SCALE)
		fprintf(outfile, CSI "%d;%ds",
		   environment[ENV_RASTER_MARGIN]/SCALE, environment[ENV_RIGHT]/SCALE);

	fprintf(outfile, "\033P0;0;%dq\n\"1;1\n", 300/environment[ENV_RASTER_RES]);
}
/*}}}*/
/*{{{  void dev_raster_data(escape_struct *escape)*/
void dev_raster_data(escape_struct *escape)
{
	int x, mask;
	char *sixel = raster_buffer;

	if (raster_width < 8*escape->param)
		raster_width = 8*escape->param;
		
	for (x=0; x<escape->param; ++x)
		for (mask=128; mask; mask >>= 1, ++sixel)
			if (escape->data[x] & mask)
				*sixel |= raster_mask;

	if ( (raster_mask <<= 1) == 64)
		flush_raster();
}
/*}}}*/
/*{{{  void dev_end_raster(escape_struct *escape)*/
void dev_end_raster(escape_struct *escape)
{
	if (raster_mask != 1)
		flush_raster();

	if (environment[ENV_RASTER_MARGIN]/SCALE != environment[ENV_LEFT]/SCALE)
		fprintf(outfile, CSI "%d;%ds",
		   environment[ENV_LEFT]/SCALE, environment[ENV_RIGHT]/SCALE);
							
	fprintf(outfile, "\x1b\\");
}
/*}}}*/

/*{{{  void dev_set_cursor(escape_struct *escape)*/
void dev_set_cursor(escape_struct *escape)
{
	fprintf(outfile, CSI "%d`" CSI "%dd", cursor->x / SCALE, cursor->y / SCALE);
}
/*}}}*/
/*{{{  void dev_set_font(int offset)*/
void dev_set_font(int offset)
{
}
/*}}}*/

/*{{{  void dev_start_string()*/
void dev_start_string()
{
}
/*}}}*/
/*{{{  void dev_write_char(int ch)*/
void dev_write_char(int ch)
{
	fputc(ch, outfile);
}
/*}}}*/
/*{{{  void dev_end_string()*/
void dev_end_string()
{
}
/*}}}*/

/*{{{  void dev_area_fill(escape_struct *escape)*/
void dev_area_fill(escape_struct *escape)
{
	pcl_area_fill(escape);
}
/*}}}*/


NULL_FN(dev_font_control)
NULL_FN(dev_select_font)
NULL_FN(dev_font_descriptor)
NULL_FN(dev_char_descriptor)
